#include <stdlib.h>
#include <stdio.h>
#include "basic.h"


static int verbose = 0 ;


/* Barrier */
#include "sense_barrier.c"


typedef struct {
  int size ;
  barrier_t *b ;
  volatile int *x,*y;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;


/* Thread 1 of test */
void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {// Main loop of size s
    wait_my_barrier(q->b,&sense) ;     //  sync with T2
    q->x[k] = 2 ; /* x <- 2 */
    q->y[k] = 1 ; /* y <- 1 */
  }
  return NULL ;
}

/* Thread 2 of test */
void *T2(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {// Main loop of size s
    wait_my_barrier(q->b,&sense) ;     //  sync with T1
    q->y[k] = 2 ; /* y <- 2 */
    q->x[k] = 1 ; /* x <- 1 */
  }
  return NULL ;
}


void run(int size) {
  barrier_t *b = alloc_my_barrier(2) ;

  for ( ; ; ) {
    /* Allocate x and y arrays of size size, initialisation to 0 */
    int *x = calloc(size,sizeof(x[0])) ;
    int *y = calloc(size,sizeof(y[0])) ;

    /* Pack test context in struct common_t */
    common_t c ;
    c.size = size ; c.b = b ; c.x = x ; c.y = y ;
    pthread_t th1,th2 ;
    ctx_t a1,a2 ;

    /* Spawn two threads for T1 and T2 */
    a1.id = 1 ; a1.common = &c ;
    create_thread(&th1,T1,&a1) ;
    a2.id = 2 ; a2.common = &c ;
    create_thread(&th2,T2,&a2) ;

    /* Wait for T1 and T2 threads to terminate */
    (void)join_thread(&th1) ;
    (void)join_thread(&th2) ;

    /* Now we read x and y and compute population counts */
    int count[3][3] ;
    count[1][1] = count[1][2] = count[2][1] = count[2][2] = 0 ;
    for (int k = 0 ; k < size ; k++) {
      int xk = x[k], yk = y[k] ;
      count[xk][yk]++ ;
    }

    /* Print population counts */
    printf("**********\n") ;
    int nz = 0 ;
    for (int k1 = 1 ; k1 <= 2 ; k1++)
      for (int k2 = 1 ; k2 <= 2 ; k2++) {
        int c = count[k1][k2] ;
        if (c > 0) nz++ ;
        printf("%-5i> x=%i y=%i\n",c,k1,k2) ;
      }

    /* Cleanup */
    free(x) ; free(y) ;

    /* If we observed four different outcomes, we are done */
    if (nz == 4) return ;
  }
  free_my_barrier(b) ;

}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 100000 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  }
  run(size) ;
  return 0 ;
}
